exec("./settings.cs");
exec("./jettison.cs");
schedule(0, 0, exec, expandFilename("./mutePackage.cs"));

$autoModRequiredMiddlewareVersion = 1.0;

if(isFile("config/server/autoModerator/apiKey.cs"))
{
	$autoModEnabled = true;
	exec("config/server/autoModerator/apiKey.cs");
}
else
{
	$autoModEnabled = false;
	schedule(30000, 0, autoModErrorAdmins, "No API key set. AutoModerator cannot run without an API key. Do /autoModeratorSetKey <API KEY> to set it. <a:github.com/conversationai/perspectiveapi/blob/master/quickstart.md>Follow these instructions until you get an API key to use</a>.");
}

if(!isObject(nameToID("chatAnalysisQueue")))
	new simGroup(chatAnalysisQueue){};

if(!isObject(nameToID("autoModScorecardGroup")))
	new simGroup(autoModScorecardGroup){};

function serverCmdAutoModeratorSetKey(%client, %key)
{
	if(!%client.isAdmin)
		return;

	if(!%client.isSuperAdmin)
	{
		%client.chatMessage("Only super-admins can update the API key.");
		return;
	}

	if(strlen(%key) < 30)
	{
		%client.chatMessage("That does not seem like a valid API key.");
		return;
	}

	$autoModWorking = false;
	$autoModPerspectiveAPIKey = %key;
	export("$autoModPerspectiveAPIKey", "config/server/autoModerator/apiKey.cs");
	deleteVariables("$autoModErrorSent*");
	%client.chatMessage("API key set. AutoModerator activated.");
	autoModMessageAdmins("\c5" @ %client.name SPC "has changed the AutoModerator API key.");
	$autoModEnabled = true;
}

function serverCmdToggleAutoModerator(%client)
{
	if(!%client.isAdmin)
		return;

	if(!%client.isSuperAdmin)
	{
		%client.chatMessage("Only super-admins can turn the AutoModerator on or off.");
		return;
	}

	$autoModEnabled = !$autoModEnabled;

	autoModMessageAdmins("\c5" @ %client.name SPC "has turned the" SPC autoModGetName() SPC ($autoModEnabled ? "on" : "off") @ ".");

	if(!$autoModEnabled)
		chatAnalysisQueue.deleteAll();
}


function chatCache::addMessage(%this, %message)
{
	if(%this.payload $= "")
	{
		%this.payload = %message;
		return;
	}

	if(strlen(%this.payload NL %message) > 3000)
	{
		%this.payloadOverflow = trim(%this.payloadOverflow NL %message);
		return;
	}

	%this.payload = trim(%this.payload NL %message);
}

function queueChatAnalysis(%message, %blid)
{
	%count = chatAnalysisQueue.getCount();
	for(%i = 0; %i < %count; %i++)
	{
		%cache = chatAnalysisQueue.getObject(%i);

		if(%cache.blid == %blid)
		{
			%cache.addMessage(%message);
			return;
		}
	}

	// Make cache for this blid
	%cache = new ScriptObject(chatCache){};
	%cache.blid = %blid;
	%cache.creationTime = $sim::time;
	%cache.addMessage(%message);

	chatAnalysisQueue.add(%cache);

	if(!isEventPending($chatAnalysisQueueSchedule))
		processChatAnalysisQueue();
}

function processChatAnalysisQueue()
{
	cancel($chatAnalysisQueueSchedule);
	$chatAnalysisQueueSchedule = schedule(1500, chatAnalysisQueue, processChatAnalysisQueue);

	%count = chatAnalysisQueue.getCount();

	if(!%count)
		return;

	for(%i = 0; %i < %count; %i++)
	{
		%cache = chatAnalysisQueue.getObject(%i);

		if(%cache.creationTime + $autoModAnalysisDelaySeconds >= $sim::time)
		{
			%cache = 0;
			continue;
		}

		break;
	}

	if(!isObject(%cache))
		return;
	
	chatAnalysisQueue.remove(%cache);
	startTextAnalysis(%cache);

	if(chatAnalysisQueue.getCount() == 0)
		cancel($chatAnalysisQueueSchedule);
}

processChatAnalysisQueue();

function startTextAnalysis(%cache)
{
	%text = %cache.payload;
	%blid = %cache.blid;

	%connection = new TCPObject(textAnalyserConnection){};
	%connection.cache = %cache;

	%post = "text=" @ urlenc(%text) @ "&key=" @ $autoModPerspectiveAPIKey;
	%httpRequest = "POST" SPC $autoModMiddlewarePath @ 
" HTTP/1.0\r\nHost: " @ $autoModMiddlewareHostname @"\r\nConnection: Close\r\nContent-Type: application/x-www-form-urlencoded\r\nContent-Length:" SPC strlen(%post) @ "\r\n\r\n" @ %post;

	%connection.blid = %blid;
	%connection.request = %httpRequest;
	%connection.connect($autoModMiddlewareHostname @ ":" @ $autoModMiddlewarePort);
	%connection.schedule(2000, onTimeout);
}

function textAnalyserConnection::onConnected(%this)
{
	%this.send(%this.request);
}

function textAnalyserConnection::onTimeout(%this)
{
	chatAnalysisQueue.add(%this.cache); // Put it back into the queue to retry analysis
	warn("Chat analyser timed out. Requeueing chat cache...");
	%this.delete();
}

function textAnalyserConnection::onLine(%this, %line)
{
	if(%pos = strpos(%line, "Content-Length: ") > -1)
	{
		%this.contentLength = mFloor(getWord(%line, 1));
		%this.lineEndings = -1;
	}

	if(!%this.readingBody && %line $= "")
	{
		%this.readingBody = true;
	}
	else if(%this.readingBody)
	{
		%this.response = %this.response @ (strlen(%this.response) > 0 ? "\r\n" : "") @ %line;
		%this.lineEndings++;
	}
}

function textAnalyserConnection::onDisconnect(%this)
{
	jettisonParse(%this.response);

	if(isObject($JSON::Value))
	{
		$autoModWorking = true;
		processAnalysisResults($JSON::Value, %this.blid);

		%cache = %this.cache;
		if(strlen(%cache.payloadOverflow) > 0)
		{
			while(strlen(%cache.payload) < 3000)
			{
				%record = getRecord(%cache.payloadOverflow, 0);

				if(strlen(%cache.payload NL %record) > 3000)
					break;

				%cache.payload = trim(%cache.payload NL %record);
				%cache.payloadOverflow = removeRecord(%cache.payloadOverflow, 0);

				chatAnalysisQueue.add(%cache); // Put it back into the queue to analyse further lines
			}
		}
		else
		{
			%cache.delete(); // This cache has served its purpose
		}
	}
	else
	{
		// error handling here....
		if(%this.response $= "ERROR: NETWORK FAILIURE")
		{
			echo("The middleware experienced a timeout.");
			chatAnalysisQueue.add(%this.cache); // Put it back into the queue to try again
			%this.delete();
			return;
		}
		else if(%this.response $= "ERROR: API RETURNED CODE 400")
		{
			if(!$autoModWorking)
				autoModErrorAdmins("The API key supplied to AutoModerator may be incorrect. Re-set the API key with /autoModeratorSetKey <API KEY>. <a:github.com/conversationai/perspectiveapi/blob/master/quickstart.md>Follow these instructions until you get a proper API key to use</a>.");
		}
		else if(%this.response $= "ERROR: API RETURNED CODE 403")
		{
			autoModErrorAdmins("The API key supplied to AutoModerator may have incorrect permission settings. Please make sure the API key is not restricted by IP address, and has the Perspective API enabled on it.");
		}
		else if(%code = getWord(%this.response, 4) > 400)
		{
			warn("API returned code" SPC %code @ "... Requeueing cache.");
			chatAnalysisQueue.add(%this.cache); // Put it back into the queue to try again
			%this.delete();
			return;
		}
		else
		{
			warn("Analyser API returned unexpected response:");
			warn(%this.response);
		}

		%this.cache.delete();
	}

	%this.delete();
}

function processAnalysisResults(%result, %blid)
{
	%version = %result.middlewareVersion;

	if(%version < $autoModRequiredMiddlewareVersion)
		autoModErrorAdmins("The middleware server configured for Server_AutoModerator is older than required. This may cause problems. Connected middleware version:" SPC %version SPC "Required:" SPC $autoModRequiredMiddlewareVersion);

	%toxicity = mFloatLength(%result.attributeScores.TOXICITY.summaryScore.value, 2);
	%severeToxicity = mFloatLength(%result.attributeScores.SEVERE_TOXICITY.summaryScore.value, 2);
	%profanity = mFloatLength(%result.attributeScores.PROFANITY.summaryScore.value, 2);
	%sexuallyExplicit = mFloatLength(%result.attributeScores.SEXUALLY_EXPLICIT.summaryScore.value, 2);
	%identityAttack = mFloatLength(%result.attributeScores.IDENTITY_ATTACK.summaryScore.value, 2);
	%flirtation = mFloatLength(%result.attributeScores.FLIRTATION.summaryScore.value, 2);
	%threat = mFloatLength(%result.attributeScores.THREAT.summaryScore.value, 2);
	%insult = mFloatLength(%result.attributeScores.INSULT.summaryScore.value, 2);

	if($autoModExperimentalMode)
		echo(%blid SPC "Toxic:" SPC mFloor(%toxicity * 100) SPC "Toxic (severe):" SPC mFloor(%severeToxicity * 100) SPC "Profanity:" SPC mFloor(%profanity * 100) SPC "Explicit:" SPC mFloor(%sexuallyExplicit * 100) SPC "ID Attack:" SPC mFloor(%identityAttack * 100) SPC "Flirt:" SPC mFloor(%flirtation * 100) SPC "Threat:" SPC mFloor(%threat * 100) SPC "Insult:" SPC mFloor(%insult * 100));

	%scorecard = 0;
	if(!isObject(%scorecard = nameToID("playerDemerits_" @ %blid)))
	{
		%scorecard = new ScriptObject("playerDemerits_" @ %blid)
		{
			blid = %blid;
			demerits["toxicity"] = 0;
			demerits["severeToxicity"] = 0;
			demerits["profanity"] = 0;
			demerits["sexuallyExplicit"] = 0;
			demerits["identityAttack"] = 0;
			demerits["flirtation"] = 0;
			demerits["threat"] = 0;
			demerits["insult"] = 0;

			punishments["toxicity"] = -1;
			punishments["severeToxicity"] = -1;
			punishments["profanity"] = -1;
			punishments["sexuallyExplicit"] = -1;
			punishments["identityAttack"] = -1;
			punishments["flirtation"] = -1;
			punishments["threat"] = -1;
			punishments["insult"] = -1;
		};

		autoModScorecardGroup.add(%scorecard);
	}

	if(%toxicity > $autoModMinimumProbability)
		%scorecard.demerits["toxicity"] += $autoModSensitivity["Toxicity"] * ((%toxicity - $autoModMinimumProbability) / (1 - $autoModMinimumProbability));
	else if(%scorecard.demerits["toxicity"] > 0)
		%scorecard.demerits["toxicity"]--;

	if(%severeToxicity > $autoModMinimumProbability)
		%scorecard.demerits["severeToxicity"] += $autoModSensitivity["severeToxicity"] * ((%severeToxicity - $autoModMinimumProbability) / (1 - $autoModMinimumProbability));
	else if(%scorecard.demerits["severeToxicity"] > 0)
		%scorecard.demerits["severeToxicity"]--;

	if(%profanity > $autoModMinimumProbability)
		%scorecard.demerits["profanity"] += $autoModSensitivity["Profanity"] * ((%profanity - $autoModMinimumProbability) / (1 - $autoModMinimumProbability));
	else if(%scorecard.demerits["profanity"] > 0)
		%scorecard.demerits["profanity"]--;

	if(%sexuallyExplicit > $autoModMinimumProbability)
		%scorecard.demerits["sexuallyExplicit"] += $autoModSensitivity["SexuallyExplicit"] * ((%sexuallyExplicit - $autoModMinimumProbability) / (1 - $autoModMinimumProbability));
	else if(%scorecard.demerits["sexuallyExplicit"] > 0)
		%scorecard.demerits["sexuallyExplicit"]--;

	if(%identityAttack > $autoModMinimumProbability)
		%scorecard.demerits["identityAttack"] += $autoModSensitivity["IdentityAttack"] * ((%identityAttack - $autoModMinimumProbability) / (1 - $autoModMinimumProbability));
	else if(%scorecard.demerits["identityAttack"] > 0)
		%scorecard.demerits["identityAttack"]--;

	if(%flirtation > $autoModMinimumProbability)
		%scorecard.demerits["flirtation"] += $autoModSensitivity["Flirtation"] * ((%flirtation - $autoModMinimumProbability) / (1 - $autoModMinimumProbability));
	else if(%scorecard.demerits["flirtation"] > 0)
		%scorecard.demerits["flirtation"]--;

	if(%threat > $autoModMinimumProbability)
		%scorecard.demerits["threat"] += $autoModSensitivity["Threat"] * ((%threat - $autoModMinimumProbability) / (1 - $autoModMinimumProbability));
	else if(%scorecard.demerits["threat"] > 0)
		%scorecard.demerits["threat"]--;

	if(%insult > $autoModMinimumProbability)
		%scorecard.demerits["insult"] += $autoModSensitivity["Insult"] * ((%insult - $autoModMinimumProbability) / (1 - $autoModMinimumProbability));
	else if(%scorecard.demerits["insult"] > 0)
		%scorecard.demerits["insult"]--;


	if(%scorecard.demerits["toxicity"] >= $autoModSensitivityLimit)
		issuePunishment(%scorecard, "toxicity");

	if(%scorecard.demerits["severeToxicity"] >= $autoModSensitivityLimit)
		issuePunishment(%scorecard, "severeToxicity");

	if(%scorecard.demerits["profanity"] >= $autoModSensitivityLimit)
		issuePunishment(%scorecard, "profanity");

	if(%scorecard.demerits["sexuallyExplicit"] >= $autoModSensitivityLimit)
		issuePunishment(%scorecard, "sexuallyExplicit");

	if(%scorecard.demerits["identityAttack"] >= $autoModSensitivityLimit)
		issuePunishment(%scorecard, "identityAttack");

	if(%scorecard.demerits["flirtation"] >= $autoModSensitivityLimit)
		issuePunishment(%scorecard, "flirtation");

	if(%scorecard.demerits["threat"] >= $autoModSensitivityLimit)
		issuePunishment(%scorecard, "threat");

	if(%scorecard.demerits["insult"] >= $autoModSensitivityLimit)
		issuePunishment(%scorecard, "insult");

	%result.delete();
}

function issuePunishment(%scorecard, %category)
{
	// Update their scorecard with this punishment
	%severity = %scorecard.punishments[%category];
	%scorecard.punishments[%category]++;
	// Partially reset their demerits in this category
	%scorecard.demerits[%category] = $autoModSensitivityLimit / 2;

	if(%severity > -1)
		%muteTime = mpow($autoModStartingMuteTime, 1 + $autoModMuteTimeExponent * %severity);
	else
		%muteTime = 0;

	%target = findClientByBL_ID(%scorecard.blid);

	if($autoModExperimentalMode)
	{
		autoModMessageAdmins("\c5[Experimental Machine Overlord] - Demerit threshhold for player" SPC %target.name SPC "(" @ %scorecard.blid @ ") exceeded. Category:" SPC %category @ ". Recommended action:" SPC (%muteTime > 0 ? "Mute" SPC %muteTime @ "s" : "Issue warning") @ ".");
		autoModMessageAdmins("\c4Note: Experimental Machine Overlord currently does not take any action and you are not obligated to follow its recommendation. Report unusual recommendations and false positives to the host. Only admins can see this message.");
	}
	else
	{
		if(%muteTime)
		{
			if(isObject(%target))
			{
				autoModMessageAdmins("\c5Mute reason:" SPC %category);
			}

			if($autoModeratorMute[%scorecard.blid] > $sim::time)
				$autoModeratorMute[%scorecard.blid] += %muteTime;
			else
				$autoModeratorMute[%scorecard.blid] = $sim::time + %muteTime;
		}
		else if(isObject(%target))
		{
			// Give a warning
			autoModMessageAdmins("\c5" @ %target.name SPC "has been warned by" SPC autoModGetName() @ ". Category:" SPC %category);
		}

		if(isObject(%target))
		{
			cancel(%target.punishmentMessageSchedule);
			%target.punishmentMessageSchedule = %target.schedule(0, statePunishmentMessage, %category);
		}
	}
}

function gameConnection::statePunishmentMessage(%this, %category)
{
	if($autoModeratorMute[%this.BL_ID] > $sim::time)
	{
		%seconds = $autoModeratorMute[%this.BL_ID] - $sim::time;
		announce("\c5" @ autoModGetName() SPC "\c6has muted\c3" SPC %this.name SPC "\c6for\c4" SPC formatSeconds(%seconds) @ "\c6.");
		warn(autoModGetName() SPC "has muted" SPC %this.name SPC "for" SPC formatSeconds(%seconds) @ ".");
	}
	else
	{
		if($autoModWarningMessage[%category] $= "")
			%message = $autoModWarningMessage["DEFAULT"];
		else
			%message = $autoModWarningMessage[%category];

		%this.chatMessage("<div:1>\c5" @ autoModGetName() @ "\c6:\c4" SPC "@" @ %this.name SPC "\c6" @ %message);

		if(isObject(nameToID("Beep_Popup_Sound")))
			%this.play2D(Beep_Popup_Sound);
	}
}

function autoModGetName()
{
	%adjectives = "Dazzling Magnificent Hideous Zealous Ambitious Fierce Mysterious Scary Colossal Massive Powerful Vast Refined Perfected Deafening Harsh Thundering Ancient Modern Acidic Spicy Tasty Freezing Incalculable Faceless";
	%adjective = getWord(%adjectives, getRandom(0, getWordCount(%adjectives) - 1));

	%name = strReplace($autoModName, "$ADJECTIVE", %adjective);

	return %name;
}

function gameConnection::displayMuteCountdown(%this)
{
	cancel(%this.muteCountdownSchedule);

	if(!%this.isTalking)
		return;

	if(isObject(nameToID("Beep_Key_Sound")))
		%this.play2D(Beep_Key_Sound);

	if($autoModeratorMute[%this.BL_ID] - $sim::time <= 1)
	{
		%this.centerPrint("\c6Your mute is over.", 6);
	}
	else
	{
		%this.centerPrint("\c6You are muted.\n\c6Your mute will end in\c4" SPC formatSeconds($autoModeratorMute[%this.BL_ID] - $sim::time) @ "\c6.", 1.5);
		%this.muteCountdownSchedule = %this.schedule(1000, displayMuteCountdown);
	}
}

function formatSeconds(%seconds)
{
	%days = mFloor(%seconds / 86400);
	%seconds -= %days * 86400;
	%hours = mFloor(%seconds / 3600);
	%seconds -= %hours * 3600;
	%minutes = mFloor(%seconds / 60);
	%seconds -= %minutes * 60;
	%seconds = mFloor(%seconds);

	if(%days)
		%string = %days SPC (%days > 1 ? "days" : "day");

	if(%hours)
		%string = %string @ (strlen(%string) > 2 ? "," : "") SPC %hours SPC (%hours == 1 ? "hour" : "hours");

	if(%minutes)
		%string = %string @ (strlen(%string) > 2 ? "," : "") SPC %minutes SPC (%minutes == 1 ? "minute" : "minutes");

	if(%seconds)
		%string = %string @ (strlen(%string) > 2 ? ", and" : "") SPC %seconds SPC (%seconds == 1 ? "second" : "seconds");

	return trim(%string);
}

function autoModErrorAdmins(%error)
{
	warn("\c0" @ %error); // Spam the console anyways lol
	%hash = sha1(%error);

	if($autoModErrorSent[%hash]) // Don't spam admins in the server if they've already seen this message once
		return;

	%count = clientGroup.getCount();
	%hits = 0;

	for(%i = 0; %i < %count; %i++)
	{
		%client = clientGroup.getObject(%i);

		if(!%client.isAdmin)
			continue;

		%hits++;
		%client.chatMessage("AutoMod Error:" SPC %error);

		if(isObject(nameToID("Beep_Popup_Sound")))
			%client.play2D(Beep_Popup_Sound);
	}

	if(%hits > 0)
		$autoModErrorSent[%hash] = true;
}

function autoModMessageAdmins(%message)
{
	warn(%message);

	for(%i = 0; %i < clientGroup.getCount(); %i++)
	{
		%client = clientGroup.getObject(%i);

		if(!%client.isAdmin)
			continue;

		%client.chatMessage(%message);

		if(isObject(nameToID("Beep_Popup_Sound")))
			%client.play2D(Beep_Popup_Sound);
	}
}

function autoModHackPackageOrder()
{
	%file = new FileObject(){};

	if(isFile("config/main.cs"))
	{
		%file.openForRead("config/main.cs");

		while(!%file.isEOF())
		{
			%line = %file.readLine();

			if(%line $= "// AUTOMOD PACKAGE EXECUTOR")
			{
				%file.close();
				%file.delete();

				if(!isPackage(autoModeratorAnalysis))
					exec("./package.cs");
				return; // Package executor is already installed, so exit here
			}
		}

		%file.close();
	}

	%file.openForAppend("config/main.cs");
	%file.writeLine("");
	%file.writeLine("");
	%file.writeLine("// AUTOMOD PACKAGE EXECUTOR");
	%file.writeLine("// If you remove this it will be recreated by Server_AutoModerator next time it is run. You might have to update this if you change the path of Server_AutoModerator.");
	%file.writeLine("exec(\"" @ expandFilename("./package.cs") @ "\");");
	%file.writeLine("if($numClientPackages $= \"\") $numClientPackages = getNumActivePackages();");
	warn("AutoModerator package preloader installed into /config/main.cs");

	%file.close();
	%file.delete();

	exec("./package.cs");
}

autoModHackPackageOrder();